/****************************************************************************/
/**    Kernel-AX25-Treiber fr GNOM.                                       **/
/**  Copyright (C) 1997 Jochen Sarrazin, DG6VJ. All righs reserved.        **/
/**                                                                        **/
/**  Ansteuerung Kernel-AX.25                                              **/
/****************************************************************************/


#include "ax25k.h"



void ax25_add_listen(char *call, char *port)
{
  s_listen *tmp;
  s_portinfo *portinfo;
  int found;
  int fd;
  struct full_sockaddr_ax25 ax25;
  int addrlen;
  
  portinfo = ax25_get_portinfo(port);

  Gross(call);
  
  /* Gucken, ob das Rufzeichen schon in der Listen-Liste dieses Ports steht */
  tmp = portinfo->listen;
  found = FALSE;
  while ((tmp != NULL) && !found)
  {
    if (!strcmp(call,tmp->call))
       found = TRUE;
    else
       tmp = tmp->next;
  }
  
  if (found) return;
  
  
  /* Nein, ist noch nicht. Eintrag machen */
  ax25.fsa_ax25.sax25_family = AF_AX25;
  ax25.fsa_ax25.sax25_ndigis = 1;
  convert_call_entry(portinfo->addr,ax25.fsa_digipeater[0].ax25_call);
  convert_call_entry(call,ax25.fsa_ax25.sax25_call.ax25_call);
  addrlen = sizeof(struct full_sockaddr_ax25);
  
  if ((fd = socket(AF_AX25,SOCK_SEQPACKET,0)) < 0) {
    syslog(LOG_ERR,"socket: %s\n",strerror(errno));
    return;
  }
  
  if (bind(fd,(struct sockaddr *)&ax25,addrlen) < 0) {
    syslog(LOG_ERR,"bind: %s\n",strerror(errno));
    close(fd);
    return;
  }

  if (listen(fd, SOMAXCONN) < 0) {
    syslog(LOG_ERR,"listen: %s\n",strerror(errno));
    close(fd);
    return;
  }
  
  
  if (portinfo->listen == NULL)
  {
    portinfo->listen = (s_listen *)malloc(sizeof(s_listen));
    tmp = portinfo->listen;
  }
  else
  {
    tmp = portinfo->listen;
    while (tmp->next != NULL) tmp = tmp->next;
    tmp->next = (s_listen *)malloc(sizeof(s_listen));
    tmp = tmp->next;
  }
  
  tmp->next = NULL;
  strcpy(tmp->call,call);
  tmp->fd = fd;
}



void ax25_delete_listen(char *call, char *port)
{
  s_listen *tmp,*last;
  s_portinfo *portinfo;
  
  
  portinfo = ax25_get_portinfo(port);
  tmp = portinfo->listen;
  last = NULL;
  while (tmp != NULL)
  {
    if (!strcmp(tmp->call,call))
    {
      close(tmp->fd);
      if (last == NULL)
         portinfo->listen = tmp->next;
      else
         last->next = tmp->next;
      free(tmp);
      return;
    }
    last = tmp;
    tmp = tmp->next;
  }
  
  
}



/* Fr dieses QSO die definierten ctl-Werte setzen. (bitte jonny, hi) */
void ax25_set_ctl(int qsonr)
{
  struct ax25_ctl_struct ax25_ctl;
  int fd;


  if (qso[qsonr]->port == NULL) return;
  
  if (convert_call_entry(qso[qsonr]->port->addr, (char *)&ax25_ctl.port_addr) == -1)
  {
    syslog(LOG_ERR,"ax25_ctl: invalid port-address");
    return;
  }
  if (convert_call_entry(qso[qsonr]->ax_call, (char *)&ax25_ctl.dest_addr) == -1)
  {
    syslog(LOG_ERR,"ax25_ctl: invalid destination-address");
    return;
  }
  if (convert_call_entry(qso[qsonr]->ax_mycall, (char *)&ax25_ctl.source_addr) == -1)
  {
    syslog(LOG_ERR,"ax25_ctl: invalid source-address");
    return;
  }
  
  
  
  if (qso[qsonr]->port->t1 != -1)
  {
    ax25_ctl.cmd = AX25_T1;
    ax25_ctl.arg = qso[qsonr]->port->t1;
  
    if (ioctl(qso[qsonr]->port->fd, SIOCAX25CTLCON, &ax25_ctl) != 0)
    {
      syslog(LOG_ERR,"ax25_ctl-error: %s",strerror(errno));
      return;
    }
  }

  if (qso[qsonr]->port->t2 != -1)
  {
    ax25_ctl.cmd = AX25_T2;
    ax25_ctl.arg = qso[qsonr]->port->t2;
  
    if (ioctl(qso[qsonr]->port->fd, SIOCAX25CTLCON, &ax25_ctl) != 0)
    {
      syslog(LOG_ERR,"ax25_ctl-error: %s",strerror(errno));
      return;
    }
  }

  if (qso[qsonr]->port->t3 != -1)
  {
    ax25_ctl.cmd = AX25_T3;
    ax25_ctl.arg = qso[qsonr]->port->t3;
  
    if (ioctl(qso[qsonr]->port->fd, SIOCAX25CTLCON, &ax25_ctl) != 0)
    {
      syslog(LOG_ERR,"ax25_ctl-error: %s",strerror(errno));
      return;
    }
  }

  if (qso[qsonr]->port->idle != -1)
  {
    ax25_ctl.cmd = AX25_IDLE;
    ax25_ctl.arg = qso[qsonr]->port->idle;
  
    if (ioctl(qso[qsonr]->port->fd, SIOCAX25CTLCON, &ax25_ctl) != 0)
    {
      syslog(LOG_ERR,"ax25_ctl-error: %s",strerror(errno));
      return;
    }
  }

  if (qso[qsonr]->port->n2 != -1)
  {
    ax25_ctl.cmd = AX25_N2;
    ax25_ctl.arg = qso[qsonr]->port->n2;
  
    if (ioctl(qso[qsonr]->port->fd, SIOCAX25CTLCON, &ax25_ctl) != 0)
    {
      syslog(LOG_ERR,"ax25_ctl-error: %s",strerror(errno));
      return;
    }
  }

  if (qso[qsonr]->port->window != -1)
  {
    ax25_ctl.cmd = AX25_WINDOW;
    ax25_ctl.arg = qso[qsonr]->port->window;
  
    if (ioctl(qso[qsonr]->port->fd, SIOCAX25CTLCON, &ax25_ctl) != 0)
    {
      syslog(LOG_ERR,"ax25_ctl-error: %s",strerror(errno));
      return;
    }
  }

  if (qso[qsonr]->port->paclen != -1)
  {
    ax25_ctl.cmd = AX25_PACLEN;
    ax25_ctl.arg = qso[qsonr]->port->paclen;
  
    if (ioctl(qso[qsonr]->port->fd, SIOCAX25CTLCON, &ax25_ctl) != 0)
    {
      syslog(LOG_ERR,"ax25_ctl-error: %s",strerror(errno));
      return;
    }
  }
  
  close(fd);
}




void ax25_set_txdelay()
{
}

void ax25_set_persis()
{
}

void ax25_set_frack()
{
}

void ax25_set_retry()
{
}

void ax25_set_maxframe()
{
}

void ax25_set_txtail()
{
}

void ax25_set_digipeat()
{
}

void ax25_set_response()
{
}

void ax25_set_idle()
{
}






/*
 * Sucht einen freien Platz im QSO-Array und allokiert den ntigen Speicher.
 */
int ax25_newqso()
{
  int i,found=FALSE;
  
  i = 0;
  while ((i < MAX_QSOs) && (!found))
  {
    if (qso[i] == NULL)
    {
      found = TRUE;
      qso[i] = (s_qsoinfo *) malloc(sizeof(s_qsoinfo));
      
      /* Variablen vorbesetzen */
      qso[i]->fd = -1;
      qso[i]->call[0] = '\0';
      qso[i]->mycall[0] = '\0';
      qso[i]->ax_call[0] = '\0';
      qso[i]->ax_mycall[0] = '\0';
      qso[i]->digis = NULL;
      qso[i]->port = NULL;

      qso[i]->saveframe = NULL;
      qso[i]->sendframe = NULL;
      
      qso[i]->status = 0;
      qso[i]->tries = 0;
      qso[i]->squeue = 0;
      qso[i]->unack = 0;
      
      qso[i]->old = FALSE;
      qso[i]->wantframes = TRUE;
    }
    else i++;
  }
  
  return(i);
}



/*
 * Gibt diese QSO-Nummer und den dazugehrigen Speicher wieder frei
 */
void ax25_freeqso(int qsonr)
{
  if (qso[qsonr] != NULL)
  {
    close(qso[qsonr]->fd);
    comm_delete_saved_data(qsonr);
    if (qso[qsonr]->digis != NULL) free(qso[qsonr]->digis);
    /* !!! qso[qsonr]->port NICHT free'en !!! */
    free(qso[qsonr]);
    qso[qsonr] = NULL;
  }
}



/*
 * Kontrolliert, ob der in bergebene port in use_ports steht. Wenn ja wird er
 * attached, ansonsten nicht beachtet
 */
int ax25_valid_port(char *port)
{
  int i;
  char tmp[500],tmp2[500];
  
  strcpy(tmp,use_ports);
  
  while (tmp[0] != '\0')
  {
    if ((i = POS(',',tmp)) == -1)
    {
      strcpy(tmp2,tmp);
      tmp[0] = '\0';
    }
    else
    {
      COPY(tmp2,tmp,0,i);
      COPY(tmp,tmp,i+1,strlen(tmp)-i-1);
    }
    
    if (!strcmp(tmp2,port)) return(TRUE);
  }
  
  return(FALSE);
}



void ax25_add_port(char *port)
{
  s_portinfo *tmp;
  char *desc;
  char *dev;
  char *addr;
  int baud;

  desc = (char *)ax25_config_get_desc(port);
  dev = (char *)ax25_config_get_dev(port);
  baud = ax25_config_get_baud(port);
  addr = (char *)ax25_config_get_addr(port);
  
  
  if (portsroot == NULL)
  {
    portsroot = (s_portinfo *)malloc(sizeof(s_portinfo));
    tmp = portsroot;
  }
  else
  {
    tmp = portsroot;
    while (tmp->next != NULL) tmp = tmp->next;
    tmp->next = (s_portinfo *)malloc(sizeof(s_portinfo));
    tmp = tmp->next;
  }
  
  tmp->next = NULL;
  tmp->name = (char *)strdup(port);
  tmp->desc = desc;
  tmp->dev = dev;
  tmp->baud = baud;
  tmp->addr = addr;
  tmp->fd = -1;

  tmp->t1 = -1;
  tmp->t2 = -1;
  tmp->t3 = -1;
  tmp->idle = -1;
  tmp->n2 = -1;
  tmp->window = -1;
  tmp->paclen = -1;
  
  tmp->listen = NULL;
}



/*
 * ffnet Sockets zu jedem konfigurierten Port
 */
void ax25_open_ports()
{
  struct full_sockaddr_ax25 ax25;
  s_portinfo *tmpport;
  int addrlen;
  
  anzports = 0;
  
  tmpport = portsroot;
  
  while (tmpport != NULL)
  {
  
    /* Socket zum Kernel-AX25 ffnen */
    if ((tmpport->fd = socket(AF_AX25, SOCK_SEQPACKET, 0)) < 0) {
      printf("port %s:\n",tmpport->name);
      printf("ERROR: socket(): %s\n", strerror(errno));
      exit(1);
    }
    
    addrlen = ax25_atonpath(&ax25,tmpport->addr);
    ax25.fsa_ax25.sax25_family = AF_AX25;
    
    if (bind(tmpport->fd, (struct sockaddr *)&ax25, addrlen) < 0) {
      printf("port %s:\n",tmpport->name);
      printf("ERROR: bind(): %s (%s)\n", strerror(errno),tmpport->addr);
      exit(1);
    }
    
    if (listen(tmpport->fd, SOMAXCONN) < 0) {
      printf("port %s:\n",tmpport->name);
      printf("ERROR: listen(): %s\n", strerror(errno));
      exit(1);
    }
    
    anzports++;
    tmpport = tmpport->next;
  }
}



/*
 * Holt Kanal-Informationen ber das Kernel-AX25 'rein
 *
 * Rckgabe: FALSE wenn Fehler, TRUE wenn kein Fehler. Fehlermeldungen schreibt die
 * Funktion selbst auf den Schirm.
 */
void ax25_load_info()
{
  char *port;
  int used_ports=0;
  
  if (ax25_config_load_ports() == 0)
  {
    printf("ERROR: No AX.25 port data configurated\n\n");
    exit(1);
  }
  
  /* Liste der Ports mit Infos dazu in die eigene Port-Liste aufnehmen */  
  port = NULL;
  do
  {
    if ((port = (char *)ax25_config_get_next(port)) != NULL)
       if (ax25_valid_port(port))
          ax25_add_port(port);
  }
  while (port != NULL);
  
}



void ax25_disconect_all()
{
  int i;
  
  while (i < MAX_QSOs)
  {
    if (qso[i] != NULL) ax25_freeqso(i);
    i++;
  }
}



void ax25_init_qsos()
{
  int i;
  
  while (i < MAX_QSOs)
  {
    qso[i] = NULL;
    i++;
  }
}



/* Gibt TRUE zurck, wenn das connectende Rufzeichen sofort wieder ohne Meldung disconnected
 * werden soll, ansonsten FALSE.
 */
int discon_iscall(char *call)
{
  s_discon *tmp;
  int found=FALSE;
  
  tmp = discroot;
  
  while ((tmp != NULL) && (found == FALSE))
  {
    if (!strcmp(tmp->call,call)) found = TRUE;
    tmp = tmp->next;
  }
  
  return(found);
}




void ax25_accept_qso(int fd, s_portinfo *port, char *mycall)
{
  int addrlen,aktqso;
  int i,k;
  struct full_sockaddr_ax25 ax25;
  char tmp[100],tmp2[100],tmp3[100];


  addrlen = sizeof(ax25);
  aktqso = ax25_newqso();
  qso[aktqso]->fd = accept(fd, (struct sockaddr *)&ax25, &addrlen);
  i = 0;
  
  /* Rufzeichen der Gegenstation */
  ax25_ntoa(qso[aktqso]->call, &ax25.fsa_ax25.sax25_call);
  /* Das Rufzeichen *MIT* SSID sichern (fr ax25_ctl) */
  strcpy(qso[aktqso]->ax_call,qso[aktqso]->call);
  
  /* SSID abhngen und testen (!= 0) */
  if ((i = POS('-',qso[aktqso]->call)) != -1)
  {
    COPY(tmp,qso[aktqso]->call,i+1,strlen(qso[aktqso]->call)-i-1);
    if (atoi(tmp) == 0) qso[aktqso]->call[i] = '\0';
  }
  
  /* Gucken, ob das Call, das connected hat, sofort disconnected werden soll */
  if (discon_iscall(qso[aktqso]->call))
  {
    ax25_freeqso(aktqso);
    return;
  }
  
  
  /* Digipeater */
  if (ax25.fsa_ax25.sax25_ndigis > 0)
  {
    k = 0;
    tmp[0] = '\0';
    while (k < ax25.fsa_ax25.sax25_ndigis)
    {
      ax25_ntoa(tmp2,&ax25.fsa_digipeater[k].ax25_call);
      
      /* SSID abhngen und testen (!= 0) */
      if ((i = POS('-',tmp2)) != -1)
      {
        COPY(tmp3,tmp2,i+1,strlen(tmp2)-i-1);
        if (atoi(tmp3) == 0) tmp2[i] = '\0';
      }
      
      if (k == 0)
         sprintf(tmp," via %s",tmp2);
      else
      {
        strcat(tmp," ");
        strcat(tmp,tmp2);
      }
      k++;
    }
    qso[aktqso]->digis = (char *)strdup(tmp);
  }
  else {
    qso[aktqso]->digis = (char *)malloc(1);
    qso[aktqso]->digis[0] = '\0';
  }
  
  /* Einen Zeiger auf die Port-Infos des Ports, auf dem connected wurde, setzen. */
  qso[aktqso]->port = port;
  
  /* Mycall */
  strcpy(qso[aktqso]->mycall,mycall);
  strcpy(qso[aktqso]->ax_mycall,qso[aktqso]->mycall);
  
  /* maxfd abndern */
  if (qso[aktqso]->fd+1 > maxfd) maxfd = qso[aktqso]->fd+1;
  
  /* Auf auf diesem QSO die in kax25.conf angegebenen Parameter setzen */
  ax25_set_ctl(aktqso);
  
  qso[aktqso]->old = FALSE;
  if (gnomfd == -1)
  {
    sprintf(tmp,"%s\r",no_gnom_text);
    write(qso[aktqso]->fd,tmp,strlen(tmp));
  }
  else
  {
    /* GNOM bescheid sagen, da jemand connected hat */
    comm_notice_connect(aktqso);
  }
}



/*
 * Es sind neue QSO-Daten gekommen - Holen und an GNOM schicken
 */
void ax25_get_qso_data(int qsonr)
{
  char tmp[1000];
  int count;
  
  if ((count = read(qso[qsonr]->fd,&tmp,999)) > 0)
  {
    comm_send_data(qsonr,tmp,count);
  }
  else
  {
    /* Den Kanal disconnecten */
    comm_notice_disconect(qsonr);
    close(qso[qsonr]->fd);		/* FD schlieen */
    ax25_freeqso(qsonr);
    return;
  }
}



/*
 * Sucht das zu dem Port gehrende Rufzeichen 'raus und gibt es zurck.
 */
int ax25_get_mycall(char *port, char *mycall)
{
  s_portinfo *tmp;
  
  tmp = portsroot;
  while (tmp != NULL)
  {
    if (!strcmp(tmp->name,port))
    {
      strcpy(mycall,tmp->addr);
      return(TRUE);
    }
    tmp = tmp->next;
  }
  
  return(FALSE);
}



s_portinfo *ax25_get_portinfo(char *port)
{
  s_portinfo *ptmp;
  
  ptmp = portsroot;
  while (ptmp != NULL)
  {
    if(!strcmp(port,ptmp->name)) return(ptmp);
    ptmp = ptmp->next;
  }
  
  return(NULL);
}



/* Wenn call_without keine SSID hat, wird -0 drangehngt, und 
 * in call_with zurck gegeben.
 */
void ax25_make_ssid_calls(char *call_with, char *call_without)
{
  int i;
  
  strcpy(call_with,call_without);
  if ((i = POS('-',call_without)) == -1)
     strcat(call_with,"-0");
}


/*
 * Connect-Versuch. Rckgabe: fd
 */
int ax25_make_connect(char *port, char *mycall, char *call, char *digis, int chan)
{
  struct full_sockaddr_ax25 ax25;
  int addrlen;
  struct full_sockaddr_ax25 ax25_d;
  int addrlen_d;
  int fd,qsonr,i;
  char tmp[500];
/*  char mycall[10];*/
  
  sprintf(tmp,"%s %s",call,digis);
  
  addrlen_d = convert_call(tmp,&ax25_d);
  
  if (addrlen_d == -1) {
    comm_connect_answer(1,0,chan);
    return(-1);
  }
  
  
  
  fd = socket(AF_AX25,SOCK_SEQPACKET,0);
  if (fd < 0) {
    comm_connect_answer(2,0,chan);
    return(-1);
  }
  
  ax25.fsa_ax25.sax25_family = AF_AX25;
  ax25.fsa_ax25.sax25_ndigis = 1;
  convert_call_entry(mycall,
                     ax25.fsa_ax25.sax25_call.ax25_call);
  convert_call_entry(ax25_config_get_addr(port),
                     ax25.fsa_digipeater[0].ax25_call);
  addrlen = sizeof(struct full_sockaddr_ax25);
  
  
  if (bind(fd,(struct sockaddr *)&ax25,addrlen) == -1) {
    close(fd);
    comm_connect_answer(3,0,chan);
    return(-1);
  }
  
  i = TRUE;
  ioctl(fd,FIONBIO,&i);
  
  if (connect(fd,(struct sockaddr *)&ax25_d,addrlen_d)) {
    if (errno != EINPROGRESS) {
      close(fd);
      comm_connect_answer(4,0,chan);
      return(-1);
    }
  }
  
  qsonr = ax25_newqso();
  qso[qsonr]->fd = fd;
  strcpy(qso[qsonr]->call,call);
  strcpy(qso[qsonr]->mycall,mycall);
  ax25_make_ssid_calls(qso[qsonr]->ax_call,call);
  ax25_make_ssid_calls(qso[qsonr]->ax_mycall,mycall);
  qso[qsonr]->digis = (char *)strdup(digis);
  qso[qsonr]->port = ax25_get_portinfo(port);
  qso[qsonr]->status = 1;
  qso[qsonr]->old = TRUE;
  /* Auf auf diesem QSO die in kax25.conf angegebenen Parameter setzen */
  ax25_set_ctl(qsonr);
  
  comm_connect_answer((char)0xFF,qsonr,chan);
  
  comm_i_want_a_frame(qsonr);
  
  return(fd);
}




void poll_ax25()
{
  char scall[10],dcall[10];
  struct proc_ax25 *p;
  struct proc_ax25 *porg;
  int i,found,axstate;
  int tries,unack,squeue;
  
  p = read_proc_ax25();
  
  if (p == NULL)
  {
    fprintf(stderr,"gnom: error reading ax25-status (read_proc_ax25)\n");
    return;
  }
  
  porg = p;
  while (p != NULL)
  {
    /* Gucken, zu welchem QSO dieser Eintrag gehrt */
    i = 0;
    found = FALSE;
    while ((i < MAX_QSOs) && (!found))
    {
      if (qso[i] != NULL)
      {
        /*
         * Hoffen wir, da es funktioniert. Gucken, da p->dest_addr und p->src_addr nicht NULL
         * sind.
         */
        if (p->dest_addr == NULL)
        {
          syslog(LOG_ERR,"p->dest_addrl == NULL!!!");
          i++;
        }
        else if (p->src_addr == NULL)
             {
               syslog(LOG_ERR,"p->src_addr == NULL!!!");
               i++;
             }
             else
             {
               strcpy(dcall,qso[i]->call);
               strcpy(scall,qso[i]->mycall);
               if (POS('-',dcall) == -1) strcat(dcall,"-0");
               if (POS('-',scall) == -1) strcat(scall,"-0");

               if ((!strcmp(p->dest_addr,dcall)) && ((!strcmp(p->src_addr,scall))))
                  found = TRUE;
               else
                  i++;
             }
      }
      else i++;
      if (found)
      {
        tries = p->n2count;
        if (p->vs < p->va)
           unack = 8 + p->vs - p->va;
        else
           unack = p->vs - p->va;
        squeue = p->sndq / 256;
        
        switch (p->st)
        {
          case 0:
          case 1:
            axstate = p->st;
            break;
          case 2:
            axstate = 3;
            break;
          case 3:
            axstate = 4;
            break;
          case 4:
            axstate = 6;
            break;
          default:
            axstate = 0;
            break;
        }
        
        if (qso[i]->status != axstate)
        {
          if ((qso[i]->status == 1) && (axstate > 1)) comm_send_saveframe(i);
          qso[i]->status = axstate;
          comm_poll_change(i,POLL_STATUS);
        }
        
        if (qso[i]->unack != unack)
        {
          qso[i]->unack = unack;
          comm_poll_change(i,POLL_UNACK);
        }
        
        if (qso[i]->tries != tries)
        {
          qso[i]->tries = tries;
          comm_poll_change(i,POLL_TRIES);
        }
        
        if (qso[i]->squeue != squeue)
        {
          comm_poll_change(i,POLL_SQUEUE);
          qso[i]->squeue = squeue;
          if ((qso[i]->squeue == 0) && (qso[i]->wantframes == FALSE))
          {
            qso[i]->wantframes = TRUE;
            comm_i_want_a_frame(i);
          }
        }
      }
    }
    p = p->next;
  }
  
  free_proc_ax25(porg);
}



