/****************************************************************************/
/**    Kernel-AX25-Treiber fr GNOM.                                       **/
/**  Copyright (C) 1997 Jochen Sarrazin, DG6VJ. All righs reserved.        **/
/**                                                                        **/
/**  Routinen fr die Kommunikation mit GNOM                               **/
/****************************************************************************/


#include "comm.h"



void comm_monitor_activate()
{
  if (monifd == -1) monitor_init();
}



void comm_monitor_deactivate()
{
  if (monifd != -1) close(monifd);
}



/*
 * Auf dieses Rufzeichen soll gehrt werden
 */
void comm_listen_to_call()
{
  int count;
  s_listen_to listen_to;
  
  if ((count = read(gnomfd,&listen_to,sizeof(s_listen_to))) != sizeof(s_listen_to)) return;
  
  ax25_add_listen(listen_to.call,listen_to.port);
}



/*
 * Auf dieses Rufzeichen nichtmehr
 */
void comm_dont_listen_to_call()
{
  int count;
  s_listen_to listen_to;
  
  if ((count = read(gnomfd,&listen_to,sizeof(s_listen_to))) != sizeof(s_listen_to)) return;
  
  ax25_delete_listen(listen_to.call,listen_to.port);
}




/*
 * Erzeugt einen Socket, und lauscht da
 */
void comm_create_and_listen_to_socket()
{
  struct sockaddr_un serv_addr;
  int servlen;
  
  if ((sockfd = socket(AF_UNIX,SOCK_STREAM,0)) < 0) {
    printf("ERROR: Can't open stream socket\n");
    exit(1);
  }
  
  bzero((char *) &serv_addr,sizeof(serv_addr));
  serv_addr.sun_family = AF_UNIX;
  strcpy(serv_addr.sun_path,socket_path);
  servlen = strlen(serv_addr.sun_path) + sizeof(serv_addr.sun_family);
  
  unlink(serv_addr.sun_path);
  
  if (bind(sockfd,(struct sockaddr *) &serv_addr,servlen) < 0) {
    printf("ERROR: Can't bind socket\n");
    exit(1);
  }
  
  
  if (listen(sockfd,4) == -1)
  {
    printf("ERROR: Can't listen to socket\n");
    exit(1);
  }
  
  chmod(socket_path,S_IWUSR|S_IRUSR|S_IXUSR|S_IWGRP|S_IRGRP|S_IXGRP|S_IWOTH|S_IROTH|S_IXOTH);


  gnomfd = -1;
}



void comm_send_driver_informations()
{
  s_driver_infos_port ptmp;
  s_portinfo *tmpports;


  /* Als erstes mal die Informationen ber den Treiber allgemein verschicken (Autor, Version etc) */
  write(gnomfd,&driver_infos_overall,sizeof(s_driver_infos_overall));
  
  /* Jetzt kommt die Anzahl der von diesem Treiber angesprochenen Kanle */
  write(gnomfd,&anzports,sizeof(anzports));
  
  /* Und jetzt die Kanal-Informationen selbst */
  tmpports = portsroot;
  while (tmpports != NULL)
  {
    strcpy(ptmp.port,tmpports->name);
    strcpy(ptmp.desc,tmpports->desc);
    ptmp.baud = tmpports->baud;
    write(gnomfd,&ptmp,sizeof(s_driver_infos_port));
    
    tmpports = tmpports->next;
  }

  /* Nachdem die Treiber-Infos abgeschickt wurden, kommen die Infos ber alle
   * andern QSOs. VORHER NICHT!!!
   */
  comm_send_allqsos();
}



/*
 * Ein disconnect-request von GNOM wird bearbeitet
 */
void comm_get_disconect_request()
{
  int qsonr;
  int count;
  
  if ((count = read(gnomfd,&qsonr,sizeof(qsonr))) != sizeof(qsonr)) return;  
  
  ax25_freeqso(qsonr);
}



/*
 * Antwort auf einen Connect-Versuch.
 * Code 1  = invalid parameter
 *      2  = cannot open socket
 *      3  = cannot bind socket
 *      4  = cannot connect twice
 *      FF = ok
 *
 * In chan wird der GNOM-Kanal zurckgegeben, damit GNOM die Zuordnung Kanal<->qsonr machen kann.
 */
void comm_connect_answer(uchar code, int qsonr, int chan)
{
  char cmd=2;
  int tmp=-1;
  
  write(gnomfd,&cmd,1);
  write(gnomfd,&code,1);
  write(gnomfd,&chan,sizeof(chan));
  if (code == 0xFF)
     write(gnomfd,&qsonr,sizeof(qsonr));
  else
     write(gnomfd,&tmp,sizeof(tmp));
}



void comm_poll_change(int qsonr, uchar what)
{
  char tmp[3];
  
  tmp[0] = 4;
  tmp[1] = what;
  switch (what)
  {
    case POLL_STATUS: tmp[2] = (char)qso[qsonr]->status; break;
    case POLL_TRIES: tmp[2] = (char)qso[qsonr]->tries; break;
    case POLL_SQUEUE: tmp[2] = (char)qso[qsonr]->squeue; break;
    case POLL_UNACK: tmp[2] = (char)qso[qsonr]->unack; break;
  }
  write(gnomfd,&tmp,3);
  write(gnomfd,&qsonr,sizeof(qsonr));
}


/*
 * von auen kommt ein connect-wunsch 'rein
 */
void comm_get_connect_request()
{
  int count;
  s_make_connect condata;
  int fd;
  
  if ((count = read(gnomfd,&condata,sizeof(condata))) != sizeof(condata)) return;
  
  fd = ax25_make_connect(condata.port,condata.mycall,condata.call,condata.digis,condata.chan);
}



/*
 * Sendet ein Frame von GNOM in Richtung Kernel-AX25 aus.
 */
void comm_get_senddata()
{
  int qsonr;
  int len;
  char *tmp;
  s_saveframe *tmp2;

  read(gnomfd,&qsonr,sizeof(qsonr));
  read(gnomfd,&len,sizeof(len));
  tmp = (char *)malloc(len);
  read(gnomfd,tmp,len);

  if (qso[qsonr]->status == 1)
  {
    /* Die Station ist noch nicht connected worden - Frame zwischenspeichern */

    if (qso[qsonr]->sendframe == NULL)
    {
      qso[qsonr]->sendframe = (s_saveframe *)malloc(sizeof(s_saveframe));
      tmp2 = qso[qsonr]->sendframe;
    }
    else
    {
      tmp2 = qso[qsonr]->sendframe;
      while (tmp2->next != NULL) tmp2 = tmp2->next;
      tmp2->next = (s_saveframe *)malloc(sizeof(s_saveframe));
      tmp2 = tmp2->next;
    }
    
    tmp2->next = NULL;
    tmp2->len = len;
    tmp2->data = (char *)malloc(len);
    memcpy(tmp2->data,tmp,len);
  }
  else write(qso[qsonr]->fd,tmp,len);

  if ((qso[qsonr]->squeue < 20) && qso[qsonr]->wantframes)
     comm_i_want_a_frame(qsonr);
  else
  {
    comm_i_dont_want_a_frame(qsonr);
    qso[qsonr]->wantframes = FALSE;
  }
}



void comm_set_txdelay()
{
  int txdelay;
  
  read(gnomfd,&txdelay,sizeof(txdelay));
  ax25_set_txdelay(txdelay);
}



/* Die Gegenseite wurde connected - alle gespeicherten Frames an den QSO-Partner
 * aussenden.
 */
void comm_send_saveframe(int qsonr)
{
  s_saveframe *tmp;
  
  while (qso[qsonr]->sendframe != NULL)
  {
    tmp = qso[qsonr]->sendframe;
    write(qso[qsonr]->fd,tmp->data,tmp->len);
    qso[qsonr]->sendframe = qso[qsonr]->sendframe->next;
    free(tmp->data);
    free(tmp);
  }
}


void comm_set_persis()
{
  int persis;
  
  read(gnomfd,&persis,sizeof(persis));
  ax25_set_persis(persis);
}


void comm_set_frack()
{
  int frack;
  
  read(gnomfd,&frack,sizeof(frack));
  ax25_set_frack(frack);
}


void comm_set_retry()
{
  int retry;
  
  read(gnomfd,&retry,sizeof(retry));
  ax25_set_retry(retry);
}


void comm_set_maxframe()
{
  int maxframe;
  
  read(gnomfd,&maxframe,sizeof(maxframe));
  ax25_set_maxframe(maxframe);
}


void comm_set_txtail()
{
  int txtail;
  
  read(gnomfd,&txtail,sizeof(txtail));
  ax25_set_txtail(txtail);
}


void comm_set_digipeat()
{
  int digipeat;
  
  read(gnomfd,&digipeat,sizeof(digipeat));
  ax25_set_digipeat(digipeat);
}


void comm_set_response()
{
  int response;
  
  read(gnomfd,&response,sizeof(response));
  ax25_set_response(response);
}


void comm_set_idle()
{
  int idle,qsonr;
  
  read(gnomfd,&qsonr,sizeof(qsonr));
  read(gnomfd,&idle,sizeof(idle));
  ax25_set_idle(qsonr,idle);
}





/*
 * Holt einen Befehl von GNOM und wertet ihn aus.
 */
void comm_check_command()
{
  int count;
  char cmd;

  /* Erstmal gucken, was das fr ein Befehl ist */
  count = read(gnomfd,&cmd,1);
  if (count == 0)
  {
    /* Lnge 0 - da hat wohl jemand GNOM beendet */
/*    ax25_disconect_all();*/
    close(gnomfd);
    gnomfd = -1;
    return;
  }

  switch (cmd)
  {
    case 1: /* Get driver informations */
            comm_send_driver_informations();
            break;
    case 2: /* connect-Request */
            comm_get_connect_request();
            break;
    case 5: /* disconnect-request */
            comm_get_disconect_request();
            break;
    case 6: /* daten in richtung kernel-ax25 schicken */
            comm_get_senddata();
            break;
    case 7: /* auf dieses Rufzeichen hren */
            comm_listen_to_call();
            break;
    case 8: /* nimmer auf dieses Rufzeichen hren */
            comm_dont_listen_to_call();
            break;
    
    /***** Hardware-Parameter einstellen **********/
    case 10: comm_set_txdelay();
             break;
    case 11: comm_set_persis();
             break;
    case 12: comm_set_frack();
             break;
    case 13: comm_set_retry();
             break;
    case 14: comm_set_maxframe();
             break;
    case 15: comm_set_txtail();
             break;
    case 16: comm_set_digipeat();
             break;
    case 17: comm_set_response();
             break;
    case 18: comm_set_idle();
             break;
             
    case 50: comm_monitor_activate();
             break;
    case 51: comm_monitor_deactivate();
             break;
  }
}



/*
 * Sagt GNOM bescheid, da jemand connectet hat
 */
void comm_notice_connect(int qsonr)
{
  s_got_connect coninfo;
  char cmd=3;

  strcpy(coninfo.port,qso[qsonr]->port->name);
  strcpy(coninfo.call,qso[qsonr]->call);
  strcpy(coninfo.path,qso[qsonr]->digis);
  strcpy(coninfo.mycall,qso[qsonr]->mycall);
  coninfo.qsonr = qsonr;
  coninfo.old = qso[qsonr]->old;
  qso[qsonr]->old = TRUE;	/* sptestens ab jetzt ist das QSO "alt" */
  
  write(gnomfd,&cmd,1);
  write(gnomfd,&coninfo,sizeof(s_got_connect));

  comm_i_want_a_frame(qsonr);
}



/*
 * GNOM bescheid sagen, da die Gegenseite disconnected hat
 */
void comm_notice_disconect(int qsonr)
{
  char cmd=5;
  
  write(gnomfd,&cmd,1);
  write(gnomfd,&qsonr,sizeof(qsonr));
}



/*
 * Diese Daten gehen in Richtung GNOM.
 *
 * Wenn GNOM nicht luft, werden die Daten lokal zwischengespeichert.
 */
void comm_send_data(int qsonr, char *data, int count)
{
  char cmd=6;
  s_saveframe *tmp;
  
  if (gnomfd == -1)
  {
    /* GNOM luft nicht. Die Daten werden in eine verkettete Liste eingetragen. */
  
    if (qso[qsonr]->saveframe == NULL)
    {
      qso[qsonr]->saveframe = (s_saveframe *)malloc(sizeof(s_saveframe));
      tmp = qso[qsonr]->saveframe;
    }
    else
    {
      tmp = qso[qsonr]->saveframe;
      while (tmp->next != NULL) tmp = tmp->next;
      tmp->next = (s_saveframe *)malloc(sizeof(s_saveframe));
      tmp = tmp->next;
    }
    
    tmp->next = NULL;
    tmp->len = count;
    tmp->data = (char *)malloc(count);
    memcpy(tmp->data,data,count);
  }
  else
  {
    /* GNOM ist da - einfach schicken */
    write(gnomfd,&cmd,1);
    write(gnomfd,&qsonr,sizeof(qsonr));
    write(gnomfd,&count,sizeof(count));
    write(gnomfd,data,count);
  }
}



/* Die Gegenseite hat disconnected, bevor die Daten abgeschickt wurden.
 * Alle Daten lschen
 */
void comm_delete_saved_data(int qsonr)
{
  s_saveframe *tmp;
  
  while (qso[qsonr]->saveframe != NULL)
  {
    tmp = qso[qsonr]->saveframe;
    qso[qsonr]->saveframe = qso[qsonr]->saveframe->next;
    free(tmp->data);
    free(tmp);
  }

  while (qso[qsonr]->sendframe != NULL)
  {
    tmp = qso[qsonr]->sendframe;
    qso[qsonr]->sendframe = qso[qsonr]->sendframe->next;
    free(tmp->data);
    free(tmp);
  }
}



/*
 * GNOM hat sich grade gemeldet. Gespeicherte Daten werden abgesendet.
 */
void comm_send_saved_data(int qsonr)
{
  s_saveframe *tmp;
  
  while (qso[qsonr]->saveframe != NULL)
  {
    tmp = qso[qsonr]->saveframe;
    comm_send_data(qsonr,tmp->data,tmp->len);
    qso[qsonr]->saveframe = qso[qsonr]->saveframe->next;
    free(tmp->data);
    free(tmp);
  }
}



/*
 * Schickt alle QSOs an GNOM.
 */
void comm_send_allqsos()
{
  int i=0;
  
  while (i < MAX_QSOs)
  {
    if (qso[i] != NULL)
    {
      comm_notice_connect(i);
      comm_send_saved_data(i);
      comm_i_want_a_frame(i);
    }
    i++;
  }

}



void comm_send_moniframe(s_monitor *moni)
{
  char cmd=50;
  
  if (gnomfd == -1) return;
  moni->anz_datalines = monidatacount;
  write(gnomfd,&cmd,1);
  write(gnomfd,moni,sizeof(s_monitor));
  while (moniroot != NULL)
  {
    write(gnomfd,&moniroot->len,sizeof(moniroot->len));
    write(gnomfd,moniroot->data,moniroot->len);
    moniroot = moniroot->next;
  }
}



void comm_i_want_a_frame(int qsonr)
{
  char cmd=40;
  
  write(gnomfd,&cmd,1);
  write(gnomfd,&qsonr,sizeof(qsonr));
}



void comm_i_dont_want_a_frame(int qsonr)
{
  char cmd=41;
  
  write(gnomfd,&cmd,1);
  write(gnomfd,&qsonr,sizeof(qsonr));
}

