/****************************************************************************/
/**    Kernel-AX25-Treiber fr GNOM.                                       **/
/**  Copyright (C) 1997 Jochen Sarrazin, DG6VJ. All righs reserved.        **/
/**                                                                        **/
/**  Routinen fr den Monitor                                              **/
/****************************************************************************/

/* Viel Code ist vom Listen-Tool der ax25-utils. Hardly umgebaut von mir (dg6vj) */

#include "comm.h"
#include <linux/if.h>
#include <linux/if_ether.h>
#include <netinet/in.h>


#define BUFSIZE		1500

#define	LAPB_UNKNOWN	0
#define	LAPB_COMMAND	1
#define	LAPB_RESPONSE	2

#define	SEG_FIRST	0x80
#define	SEG_REM		0x7F

#define	PID_SEGMENT	0x08
#define	PID_ARP		0xCD
#define	PID_NETROM	0xCF
#define	PID_IP		0xCC
#define	PID_X25		0x01
#define	PID_TEXNET	0xC3
#define	PID_FLEXNET	0xCE
#define	PID_NO_L3	0xF0

#define	I		0x00
#define	S		0x01
#define	RR		0x01
#define	RNR		0x05
#define	REJ		0x09
#define	U		0x03
#define	SABM		0x2F
#define	SABME		0x6F
#define	DISC		0x43
#define	DM		0x0F
#define	UA		0x63
#define	FRMR		0x87
#define	UI		0x03
#define	PF		0x10
#define	EPF		0x01

#define	MMASK		7

#define	HDLCAEB		0x01
#define	SSID		0x1E
#define	REPEATED	0x80
#define	C		0x80
#define	SSSID_SPARE	0x40
#define	ESSID_SPARE	0x20

#define	ALEN		6
#define	AXLEN		7

#define	W		1
#define	X		2
#define	Y		4
#define	Z		8


static int  ftype(unsigned char *, int *, int *, int *, int *, int);
static char *decode_type(int);
static char *pax25(char *buf, unsigned char *data);
static void readable_dump(unsigned char *data, int length);


/* Daten dieses Monitor-Frames */
s_monitor moni;


void monitor_init()
{
  monifd = -1;
  
  
  if ((monifd = socket(AF_INET, SOCK_PACKET, htons(proto))) == -1)
  {
    syslog(LOG_ERR,"monitor_init: socket() - failure.");
    return;
  }
  
}



void ax25_dump(unsigned char *data, int length)
{
  char tmp[15];
  int ctlen, nr, ns, pf, pid, seg, type, end, cmdrsp, extseq,i;
  char *dama;

  /* check for FlexNet compressed header first; FlexNet header compressed packets are at least 8 bytes long */
  if (length < 8)
  {
    /* Something wrong with the header */
    moni.flags |= MONI_BAD_HEADER;
    return;
  }
  
  if (data[1] & HDLCAEB)
  {
    /* this is a FlexNet compressed header */
    tmp[6] = tmp[7] = extseq = 0;
    tmp[0] = ' ' + (data[2] >> 2);
    tmp[1] = ' ' + ((data[2] << 4) & 0x30) + (data[3] >> 4);
    tmp[2] = ' ' + ((data[3] << 2) & 0x3c) + (data[4] >> 6);
    tmp[3] = ' ' + (data[4] & 0x3f);
    tmp[4] = ' ' + (data[5] >> 2);
    tmp[5] = ' ' + ((data[5] << 4) & 0x30) + (data[6] >> 4);
    if (data[6] & 0xf)
       sprintf(tmp+7, "-%d", data[6] & 0xf);
    sprintf(moni.calltocall,"%d->%s%s",(data[0] << 6) | ((data[1] >> 2) & 0x3f), strtok(tmp," "),tmp+7);
    cmdrsp = (data[1] & 2) ? LAPB_COMMAND : LAPB_RESPONSE;
    data   += 7;
    length -= 7;
  }
  else
  {
    /* Extract the address header */
    if (length < (AXLEN + AXLEN + 1))
    {
      /* Something wrong with the header */
      moni.flags |= MONI_BAD_HEADER;
      return;
    }
		
    if ((data[AXLEN + ALEN] & SSSID_SPARE) == SSSID_SPARE)
       extseq = 0;
    else
    {
      extseq = 1;
      moni.flags |= MONI_AX_EAX;
    }
    
    if ((data[AXLEN + ALEN] & ESSID_SPARE) != ESSID_SPARE)
       moni.flags |= MONI_DAMA;
    
    sprintf(moni.calltocall,"%s->",pax25(tmp, data + AXLEN));
    strcat(moni.calltocall,pax25(tmp, data));
    
    cmdrsp = LAPB_UNKNOWN;
    
    if ((data[ALEN] & C) && !(data[AXLEN + ALEN] & C))
       cmdrsp = LAPB_COMMAND;
    
    if ((data[AXLEN + ALEN] & C) && !(data[ALEN] & C))
       cmdrsp = LAPB_RESPONSE;
    
    end = (data[AXLEN + ALEN] & HDLCAEB);
    
    data   += (AXLEN + AXLEN);
    length -= (AXLEN + AXLEN);

    if (!end)
    {
      i = 0;
      while (!end)
      {
        /* Print digi string */
        if (i == 0)
           sprintf(moni.digis,"%s%s%s",moni.digis, pax25(tmp, data), (data[ALEN] & REPEATED) ? "*" : "");
        else
           sprintf(moni.digis,"%s %s%s",moni.digis, pax25(tmp, data), (data[ALEN] & REPEATED) ? "*" : "");
        i++;
          
        end = (data[ALEN] & HDLCAEB);
          
        data   += AXLEN;
        length -= AXLEN;
      }
    }
  }

  if (length == 0) return;

  ctlen = ftype(data, &type, &ns, &nr, &pf, extseq);
	
  data   += ctlen;
  length -= ctlen;

  switch (type)
  {
    case I: moni.frametype = 3; break;
    case SABM: moni.frametype = 1; break;
    case SABME: moni.frametype = 11; break;
    case DISC: moni.frametype = 7; break;
    case DM: moni.frametype = 8; break;
    case UA: moni.frametype = 2; break;
    case RR: moni.frametype = 4; break;
    case RNR: moni.frametype = 5; break;
    case REJ: moni.frametype = 12; break;
    case FRMR: moni.frametype = 9; break;
    case UI: moni.frametype = 10; break;
    default: moni.frametype = -1; break;
  }
    
  if (pf) moni.flags |= MONI_POLLFINAL;
  
  switch (cmdrsp)
  {
    case LAPB_COMMAND:
         moni.cmd_resp = 1;
         break;
    case LAPB_RESPONSE:
         moni.cmd_resp = 2;
    default:
         moni.cmd_resp = 0;
         break;
  }
    
  if (type == I)
     moni.txcount = ns;
    
  if ((type & 0x3) != U)   /* I or S frame? */
     moni.rxcount = nr;
    
  if (type == I || type == UI)
  {
    /* Decode I field */
    if (length > 0)
    {        /* Get pid */
      pid = *data++;
      length--;
      
/*      if (pid == PID_SEGMENT)
      {
        seg = *data++;
        length--;
        sprintf(this_string, "%s%s remain %u",this_string, seg & SEG_FIRST ? " First seg;" : "", seg & SEG_REM);
  
        if (seg & SEG_FIRST)
        {
          pid = *data++;
          length--;
        }
      }*/


      moni.pid = pid;
      switch (pid)
      {
/*        case PID_SEGMENT:
             data_dump(data, length, hexdump);
             break;
        case PID_ARP:
             arp_dump(data, length);
             break;
        case PID_NETROM:
             netrom_dump(data, length, hexdump);
             break;
        case PID_IP:
             ip_dump(data, length, hexdump);
             break;
        case PID_X25:
             rose_dump(data, length, hexdump);
             break;
        case PID_TEXNET:
             data_dump(data, length, hexdump);
             break;
        case PID_FLEXNET:
             flexnet_dump(data, length, hexdump);
             break;
        case PID_NO_L3:
/*             data_dump(data, length, hexdump);
             break;*/
        default:
             readable_dump(data, length);
             break;
      }
    }
  }
  else
    if (type == FRMR && length >= 3)
    {
      /* FIX ME XXX
         lprintf(T_AXHDR, ": %s", decode_type(ftype(data[0])));
       */
      sprintf(moni.frmr_info, "%02X", data[0]);
      sprintf(moni.frmr_info, "%s Vr = %d Vs = %d",moni.frmr_info,
                              (data[1] >> 5) & MMASK,
			      (data[1] >> 1) & MMASK);
      if(data[2] & W)
         strcat(moni.frmr_info, " Invalid control field");
      if(data[2] & X)
         strcat(moni.frmr_info, " Illegal I-field");
      if(data[2] & Y)
         strcat(moni.frmr_info, " Too-long I-field");
      if(data[2] & Z)
         strcat(moni.frmr_info, " Invalid seq number");
      sprintf(moni.frmr_info,"%s%s",moni.frmr_info, dama);
    }
/*    else
    if ((type == SABM || type == UA) && length >= 2)
    {
      /* FlexNet transmits the QSO "handle" for header
       * compression in SABM and UA frame data fields 
       */
/*      sprintf(this_string,"%s [%d]%s",this_string, (data[0] << 8) | data[1], dama);
      sendline();
    }
    else
    {
      sprintf(this_string,"%s%s",this_string, dama);
      sendline();
    }*/
}



#define	PARAM_DATA	0
#define	PARAM_TXDELAY	1
#define	PARAM_PERSIST	2
#define	PARAM_SLOTTIME	3
#define	PARAM_TXTAIL	4
#define	PARAM_FULLDUP	5
#define	PARAM_HW	6
#define	PARAM_RETURN	15		/* Should be 255, but is ANDed with 0x0F */


void kiss_dump(unsigned char *data, int length)
{
  int type;

  type = data[0] & 0xf;
  
  if (type == PARAM_DATA)
  {
    ax25_dump(data + 1, length - 1);
    return;
  }
}



void monitor_operate()
{
  int count,found=FALSE;
  char temp[500],portname[50];
  struct sockaddr sa;
  int asize = sizeof(sa);
  s_portinfo *port;
  struct ifreq ifr;
  unsigned char buffer[BUFSIZE];
  

  /* Den Inhalt der moni-struktur leeren */
  moni.port[0] = '\0';
  moni.flags = 0;
  moni.calltocall[0] = '\0';
  moni.digis[0] = '\0';
  moni.frametype = 0;
  moni.cmd_resp = 0;
  moni.pid = 0;
  moni.rxcount = -1;
  moni.txcount = -1;
  moni.anz_datalines = 0;
  
  moniroot = NULL;
  monidatacount = 0;
  
  

  if ((count = recvfrom(monifd, buffer, sizeof(buffer), 0, &sa, &asize)) == -1)
  {
    syslog(LOG_ERR,"monitor_operate: recvfrom() - failure.");
    return;
  }
  
  /* Isses ein von uns beobachteter Port? */
  strcpy(portname,sa.sa_data);
  port = portsroot;
  while ((port != NULL) && (!found))
  {
    if (!strcmp(sa.sa_data,port->dev))
    {
      found = TRUE;
      strcpy(portname,port->name);
    }
    else
       port = port->next;
  }
  
  if (!found) return;
  
  if (proto == ETH_P_ALL)
  {
    strcpy(ifr.ifr_name, sa.sa_data);
    if (ioctl(monifd, SIOCGIFHWADDR, &ifr) < 0)
      syslog(LOG_ERR,"monitor_operate: ioctl(GIFADDR) - failure.");

    if (ifr.ifr_hwaddr.sa_family == AF_AX25)
    {
      strcpy(moni.port,portname);
      kiss_dump(buffer, count);
      comm_send_moniframe(&moni);
    }
  }
  else
  {
    strcpy(moni.port,portname);
    kiss_dump(buffer, count);
    comm_send_moniframe(&moni);
  }
}



static char *pax25(char *buf, unsigned char *data)
{
	int i, ssid;
	char *s;
	char c;
	
	s = buf;
	
	for (i = 0; i < ALEN; i++) {
		c = (data[i] >> 1) & 0x7F;

		if (!isalnum(c) && c != ' ') {
			strcpy(buf, "[invalid]");
			return buf;
		}
		
		if (c != ' ') *s++ = c;
	}	

	if ((ssid = (data[ALEN] & SSID)) != 0)
		sprintf(s, "-%d", ssid >> 1);
	else
		*s = '\0';

	return(buf);
}

static int ftype(unsigned char *data, int *type, int *ns, int *nr, int *pf, int extseq)
{
	if (extseq) {
		if ((*data & 0x01) == 0) {	/* An I frame is an I-frame ... */
			*type = I;
			*ns   = (*data >> 1) & 127;
			data++;
			*nr   = (*data >> 1) & 127;
			*pf   = *data & EPF;
			return 2;
		}
		if (*data & 0x02) {
			*type = *data & ~PF;
			*pf   = *data & PF;
			return 1;
		} else {
			*type = *data;
			data++;
			*nr   = (*data >> 1) & 127;
			*pf   = *data & EPF;
			return 2;
		}
	} else {
		if ((*data & 0x01) == 0) {	/* An I frame is an I-frame ... */
			*type = I;
			*ns   = (*data >> 1) & 7;
			*nr   = (*data >> 5) & 7;
			*pf   = *data & PF;
			return 1;
		}
		if (*data & 0x02) {		/* U-frames use all except P/F bit for type */
			*type = *data & ~PF;
			*pf   = *data & PF;
			return 1;
		} else {			/* S-frames use low order 4 bits for type */
			*type = *data & 0x0F;
			*nr   = (*data >> 5) & 7;
			*pf   = *data & PF;
			return 1;
		}
	}
}




static void readable_dump(unsigned char *data, int length)
{
  unsigned char c;
  int  i;
  int  cr = 1;
  char buf[BUFSIZE];

  for (i = 0; length > 0; i++)
  {
    c = *data++;
    length--;
    
    switch (c)
    {
      case 0x00:
           buf[i] = ' ';
      case 0x0A: /* hum... */
      case 0x0D:
           if (cr)
              buf[i] = '\xD';
           else
              i--;
           break;
      default:
           buf[i] = c;
    }
    cr = (buf[i] != '\xD');
  }
  if (cr)
    buf[i++] = '\xD';
  buf[i++] = '\0';
  
  moniroot = (s_monidata *) malloc(sizeof(s_monidata));
  moniroot->data = (char *) malloc(i);
  moniroot->len = i-1;
  memcpy(moniroot->data,buf,i);
  moniroot->next = NULL;
  monidatacount = 1;
}



